package com.fjhx.tenant.utils;

import cn.hutool.core.util.ObjectUtil;
import cn.hutool.extra.spring.SpringUtil;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.ruoyi.common.core.domain.entity.SysDept;
import com.ruoyi.common.exception.ServiceException;
import com.ruoyi.system.service.ISysDeptService;

import java.util.*;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Collectors;

public class DeptUtils {

    private static final ISysDeptService sysDeptService = SpringUtil.getBean(ISysDeptService.class);

    /**
     * 根据部门id获取公司信息
     */
    public static SysDept getCompanyByDeptId(Long deptId) {
        SysDept dept = sysDeptService.getById(deptId);
        if (ObjectUtil.isEmpty(dept)) {
            throw new ServiceException("查询不到该部门公司信息");
        }
        if (Objects.equals(dept.getType(), 0)) {
            return dept;
        }
        return getCompanyByDeptId(dept.getParentId());
    }

    public static Long getCompanyIdByDeptId(Long deptId) {
        return getCompanyByDeptId(deptId).getDeptId();
    }

    /**
     * 根据部门id列表获取部门名称
     */
    public static Map<Long, String> getDeptNameMap(List<Long> deptIds) {
        List<SysDept> deptList = sysDeptService.list(Wrappers.<SysDept>query()
                .in(ObjectUtil.isNotEmpty(deptIds), "dept_id", deptIds)
        );
        return deptList.stream().collect(Collectors.toMap(SysDept::getDeptId, SysDept::getDeptName));
    }

    public static Map<Long, String> getDeptNameMap() {
        List<SysDept> deptList = sysDeptService.list();
        return deptList.stream().collect(Collectors.toMap(SysDept::getDeptId, SysDept::getDeptName));
    }

    public static <T> void assignmentNickName(List<T> list, Function<T, Long> idMapper, BiConsumer<T, String> setNickNameMap) {
        if (list == null || list.size() == 0) {
            return;
        }
        Map<Long, String> nickNameMapByIds = getDeptNameMap();
        list.forEach(item -> setNickNameMap.accept(item, nickNameMapByIds.get(idMapper.apply(item))));
    }

    /**
     * 赋值部门名称列表树
     */
    public static <T> void assignmentDeptNames(List<T> dataList, Function<T, Long> getIdFun, BiConsumer<T, List<String>> setAttributeFun) {
        Map<Long, SysDept> sysDeptMap = sysDeptService.list().stream().collect(Collectors.toMap(SysDept::getDeptId, Function.identity()));
        for (T item : dataList) {
            //赋值部门信息
            List<String> deptNameList = new ArrayList<>();
            Long deptId = getIdFun.apply(item);
            while (true) {
                SysDept sysDept = sysDeptMap.get(deptId);
                if (ObjectUtil.isEmpty(sysDept)) {
                    break;
                }
                deptNameList.add(sysDept.getDeptName());
                if (sysDept.getType() == 0) {
                    break;
                }
                deptId = sysDept.getParentId();
            }
            //递归完数据是反的翻转一下
            Collections.reverse(deptNameList);
            setAttributeFun.accept(item, deptNameList);
        }
    }

}
